/*******************************************************************************
** Patcher (c) 2007 Steven H. Don                                             **
**                                                                            **
** Simple program to create patches for binary files or to apply those        **
** patches. Includes the ability to embed the generated patch and this        **
** program itself in an executable form for easy distribution of patches.     **
** For questions, feel free to e-mail me.                                     **
**                                                                            **
** This program was designed to compile on Win32 as well as on Linux and      **
** probably other operating systems, too, as long as they can compile         **
** standard C++ code. The code has been tested with Microsoft Visual C++ 6    **
** and 7/2003 as well as GCC 4.1                                              **
**                                                                            **
**    shd@earthling.net                                                       **
**    http://www.shdon.com/                                                   **
**                                                                            **
********************************************************************************
** Version history:                                                           **
**                                                                            **
** 2007-07-13 - 1.0 - Initial release                                         **
** 2011-07-03 - 1.1 - Bugfix to enable "noskipws" flag on input streams to    **
**                    prevent desynchronisation of input streams or corrupt   **
**                    patches.                                                **
** 2011-07-10 - 1.2 - Minor update to show filesize in decimal, rather than   **
**                    hexadecimal and to show a message about embedding the   **
**                    patch in an executable, rather than showing only the    **
**                    temporary filename.                                     **
*******************************************************************************/

#include <fstream>
#include <iomanip>
#include <iostream>
#include <list>
#include <string>
#include <vector>

#include <cstdio>

#ifdef WIN32
#include <windows.h>
#endif

// Type definitions
////////////////////////////////////////////////////////////////////////////////

const char SIGNATURE [] = "PATCH1";

enum ERRCODE
{
	ERR_INVALID_PARAM = 1,
	ERR_NO_PATCH,
	ERR_MISSING_SOURCE,
	ERR_CANT_READ,
	ERR_CANT_WRITE,
	ERR_SIZE_MISMATCH,
	ERR_IDENTICAL,
	ERR_INVALID_HEADER,
	ERR_ALREADY_PATCHED,
	ERR_DATA_MISMATCH
};

typedef std::vector <unsigned char> BYTES;
typedef BYTES::const_iterator BYTES_CITOR;
typedef BYTES::reverse_iterator BYTES_RITOR;

struct SDifference
{
	//Default constructor does nothing
	SDifference () {}

	//Copy constructor, actually swaps around
	SDifference (const SDifference &rhs)
	{
		*this = rhs;
	}

	//Assignment operator, actually swaps around
	void operator = (const SDifference &rhs)
	{
		/*
			Although rhs is supposedly const, it does change here. It shouldn't
			have been declared as const, but std::list::push_back expects a
			const parameter and looks for a copy constructor with a const
			parameter.

			Since this program discards the original after copying, it doesn't
			matter that the original changes. To be able to change it, the
			byte sequences src and tgt must be declared "mutable".
		*/
		offset = rhs.offset;
		src.swap (rhs.src);
		tgt.swap (rhs.tgt);
		trim ();
	}

	void add (unsigned char a, unsigned char b)
	{
		src.push_back (a);
		tgt.push_back (b);
	}

	void trim () const
	{
		//Set up iterators
		BYTES_RITOR itor_src = src.rbegin ();
		BYTES_RITOR itor_tgt = tgt.rbegin ();
		BYTES_RITOR end = src.rend ();

		//Find number of identical trailing bytes
		int identical = 0;
		while (itor_src != end)
		{
			if (*itor_src++ == *itor_tgt++) ++identical;
			else break;
		}

		//Remove them
		if (identical)
		{
			src.resize (src.size () - identical);
			tgt.resize (tgt.size () - identical);
		}
	}

	size_t offset;
	mutable BYTES src;
	mutable BYTES tgt;
};

typedef std::list <SDifference> DIFFS;
typedef DIFFS::iterator DIFFS_ITOR;

// Operator overloads for serialisation
////////////////////////////////////////////////////////////////////////////////

/**
 * Serialize an SDifference struct to an output stream.
 *
 * @param os The stream to output to.
 * @param diff The struct to output.
 */
std::ostream &operator << (std::ostream &os, const SDifference &diff)
{
	//Output offset
	os.write ((const char *)&diff.offset, sizeof (diff.offset));

	//Output size
	size_t size = (size_t)diff.src.size ();
	os.write ((const char *)&size, sizeof (size));

	//Output original bytes
	os.write ((const char *)&diff.src [0], size);

	//Output patched bytes
	os.write ((const char *)&diff.tgt [0], size);

	//All done
	return os;
}

/**
 * Unserialize an SDifference struct from an input stream.
 *
 * @param os The stream to read from.
 * @param diff The struct to read to.
 */
std::istream &operator >> (std::istream &is, SDifference &diff)
{
	//Read offset
	is.read ((char *)&diff.offset, sizeof (diff.offset));

	//Read size
	size_t size = 0;
	is.read ((char *)&size, sizeof (size));

	//Read original bytes
	diff.src.resize (size);
	is.read ((char *)&diff.src [0], size);

	//Read patched bytes
	diff.tgt.resize (size);
	is.read ((char *)&diff.tgt [0], size);

	//All done
	return is;
}

/**
 * Serialize an std::string to a file. Outputs the length first, then the actual
 * character data.
 *
 * @param os The stream to output to.
 * @param str The string to output.
 */
std::ofstream &operator << (std::ofstream &os, const std::string &str)
{
	size_t size = (size_t)str.length ();
	os.write ((const char *)&size, sizeof (size));
	os.write (str.c_str (), size);
	return os;
}

/**
 * Unserialize an std::string from a file. Reads the length first, prepares the
 * string, then reads the actual character data.
 *
 * @param is The stream to read from.
 * @param str The string to read to.
 */
std::ifstream &operator >> (std::ifstream &is, std::string &str)
{
	size_t size = 0;
	is.read ((char *)&size, sizeof (size));
	str.resize (size);
	is.read (&str [0], size);
	return is;
}

// Internal functions
////////////////////////////////////////////////////////////////////////////////

/**
 * Print an error message and end the program.
 *
 * @param code The numerical error code.
 * @param msg An error string.
 */
void die (ERRCODE code, const char *msg)
{
	//Dump error message
	std::cerr << "Error: " << msg << std::endl;
	exit (code);
}

void embed_patch (const std::string &execname, const std::string &patch, const std::string &output)
{
	std::ifstream e (execname.c_str (), std::ios::binary);
	if (!e) die (ERR_CANT_READ, "Cannot use this executable to build embedded patch");

	std::ifstream p (patch.c_str (), std::ios::binary);
	if (!p) die (ERR_CANT_READ, "Cannot read generated patch from temporary file");

	std::ofstream o (output.c_str (), std::ios::binary);
	if (!o) die (ERR_CANT_WRITE, "Cannot write to output file");

	std::cout << "Embedding patch into executable: " << output << std::endl;

	char buf [1024];
	while (!e.eof ())
	{
		e.read (buf, sizeof (buf));
		o.write (buf, e.gcount ());
	}
	size_t offset = size_t (o.tellp ());
	while (!p.eof ())
	{
		p.read (buf, sizeof (buf));
		o.write (buf, p.gcount ());
	}
	o.write ((const char *)&offset, sizeof (offset));
}

/**
 * Checks the executable file for an embedded patch. If it is found, this
 * is written to a temporary file for which the name is returned.
 *
 * @param execname The filename of the executable file.
 */
const std::string &get_embedded (const std::string &execname)
{
	static std::string result;

	//Open the executable and check the end for an offset
	std::ifstream e (execname.c_str (), std::ios::binary);
	if (e)
	{
		//Get the offset
		size_t offset = 0;
		e.seekg (-(int)sizeof (offset), std::ios::end);
		size_t end = (size_t)e.tellg ();
		e.read ((char *)&offset, sizeof (offset));

		//Must be in front of the offset data itself
		if (offset > 0 && offset < end)
		{
			e.seekg (offset);

			//Check for signature
			char hdr [] = "patch*";
			e.read (hdr, strlen (hdr));
			if (!strcmp (hdr, SIGNATURE))
			{
				//Found it, so extract it
				char *tmp = tmpnam (NULL);
				if (tmp)
				{
					//Open temporary file
					std::ofstream t (tmp, std::ios::binary | std::ios::trunc);
					if (t)
					{
						//Load from executable
						BYTES data;
						data.resize (end - offset);
						e.seekg (offset);
						e.read ((char *)&data [0], data.size ());

						//Write to new file
						t.write ((const char *)&data [0], data.size ());

						//All done, so return the filename
						result = tmp;
					}
				}
			}
		}
	}

	return result;
}

/**
 * Retrieves the filename of this executable.
 *
 * @param arg0 The first parameter passed to main.
 */
const std::string &get_running (const char *arg0)
{
	static std::string execname = arg0;

#ifdef WIN32
	/* WIN32 specific: first argument may not be the file name. Get it properly */
	HMODULE hModule = GetModuleHandle (NULL);
	DWORD nSize = MAX_PATH;
	for (;;)
	{
		execname.resize (nSize);

		DWORD nWritten = GetModuleFileName (hModule, &execname [0], execname.size ());
		if (nWritten < nSize)
		{
			execname.resize (nWritten);
			break;
		}

		//Increase the size of the buffer and try again
		nSize <<= 1;
	}
#endif

	return execname;
}

/**
 * Display the specified string to the user. On Windows systems, this will be
 * opened in NOTEPAD (or whatever the standard text editor is). If that is not
 * possible, it will be copied to STDOUT.
 *
 * @param str The message to display
 */
void present_message (const std::string &str)
{
	bool displayed = false;

#ifdef WIN32
	/* WIN32 specific: Write to temporary file and open in text editor */

	//Set up for command building
	const char START [] = "START ";
	const int LEN_START = strlen (START);
	const char FILENAME [] = "PATCH.TXT";
	const int LEN_FILENAME = strlen (FILENAME);

	//Get size of buffer for temporary directory
	DWORD nBufferLength = GetTempPath (0, NULL);

	//Prepare buffer
	char *lpCommand = new char [nBufferLength + LEN_START + LEN_FILENAME];

	//Fill it with the start command
	strcpy (lpCommand, START);

	//Append the temporary directory
	char *lpTemp = lpCommand + LEN_START;
	if (GetTempPath (nBufferLength, lpTemp) <= nBufferLength)
	{
		//Append the filename
		strcat (lpTemp, FILENAME);

		//Write to the temporary file
		HANDLE hFile = CreateFile (lpTemp, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
		if (hFile != INVALID_HANDLE_VALUE)
		{
			DWORD nNumberOfBytesWritten = 0;
			WriteFile (hFile, str.c_str (), str.size (), &nNumberOfBytesWritten, NULL);
			CloseHandle (hFile);
			if (nNumberOfBytesWritten == str.size ())
				displayed = (system (lpCommand) != -1);
		}
	}
#endif

	//Output to console if not already displayed
	if (!displayed) std::cout << str << std::endl;
}

// Main code
////////////////////////////////////////////////////////////////////////////////

/**
 * Apply the specified patch. Opens the patch file, reads the instructions from
 * it and carries them out.
 *
 * @param patch A string containing a file name for the patch to read.
 * @param reverse If true, the patch will be reversed in memory before being applied (basically undoes the patch).
 */
void apply_patch (const std::string &patch, bool reverse)
{
	std::cout << "Reading patch file: " << patch << std::endl;

	//Attempt to open the files to compare and immediately seek to the end
	std::ifstream p (patch.c_str (), std::ios::binary);
	if (!p) die (ERR_CANT_READ, "Could not open patch file");

	//Check header
	char hdr [] = "patch*";
	p.read (hdr, strlen (hdr));
	if (strcmp (hdr, SIGNATURE)) die (ERR_INVALID_HEADER, "Not a valid patch file");

	//Get original filename and file size
	std::string original;
	p >> original;
	size_t size;
	p.read ((char *)&size, sizeof (size));

	//Read and output description
	std::string msg;
	p >> msg;
	if (!msg.empty ()) present_message (msg);

	//Read differences from file
	DIFFS diffs;
	for (;;)
	{
		SDifference diff;
		p >> diff;
		if (p.eof ()) break;
		diffs.push_back (diff);
	}

	//Reverse the patch data in memory, if necessary
	if (reverse)
	{
		std::cout << "Reversing patch data...";

		DIFFS_ITOR itor = diffs.begin ();
		DIFFS_ITOR end = diffs.end ();
		while (itor != end)
		{
			SDifference &diff = *itor++;
			diff.src.swap (diff.tgt);
		}

		std::cout << " DONE" << std::endl;
	}

	//Output info about the patch
	std::cout << "Patch for file " << original << " (" << size << " bytes)" << std::endl;

	//Attempt to open the file
	std::fstream f (original.c_str (), std::ios::binary | std::ios::ate | std::ios::in | std::ios::out);
	if (!f) die (ERR_CANT_READ, "Could not open file to patch");

	//Check the size
	if ((size_t)f.tellg () != size) die (ERR_SIZE_MISMATCH, "File does not match size recorded in patch");

	//Check whether the data is correct
	bool match_src = true;
	bool match_tgt = true;
	{
		DIFFS_ITOR itor = diffs.begin ();
		DIFFS_ITOR end = diffs.end ();
		while (itor != end)
		{
			SDifference &diff = *itor++;

			//Seek to the correct location in the file
			f.seekg (diff.offset);

			//Read the appropriate section from the file
			BYTES ref;
			ref.resize (diff.src.size ());
			f.read ((char *)&ref [0], ref.size ());

			//Does it match the source or the destination?
			if (match_src && memcmp (&ref [0], &diff.src [0], ref.size ())) match_src = false;
			if (match_tgt && memcmp (&ref [0], &diff.tgt [0], ref.size ())) match_tgt = false;
			if (!match_src && !match_tgt) break;
		}
	}

	//Can the patch be applied?
	if (match_tgt) die (ERR_ALREADY_PATCHED, "File is already patched, specify \"--reverse\" to undo");
	if (!match_src) die (ERR_DATA_MISMATCH, "Patch does not apply to this file (data mismatch)");

	//Apply the patch
	std::cout << "Applying patch:" << std::endl;
	{
		DIFFS_ITOR itor = diffs.begin ();
		DIFFS_ITOR end = diffs.end ();
		while (itor != end)
		{
			SDifference &diff = *itor++;

			//Seek to the correct location in the file
			f.seekg (diff.offset);

			//Write the appropriate section to the file
			f.write ((const char *)&diff.tgt [0], diff.tgt.size ());

			//Display progress
			std::cout.setf (std::ios::uppercase);
			std::cout
				<< "@ " << std::setbase (16) << std::setw (8) << std::setfill ('0') << diff.offset
				<< " : " << std::setbase (10) << diff.tgt.size () << " bytes" << std::endl;
		}
	}
}

/**
 * Create a new patch. Opens the original and patched files, compares them and
 * writes the differences to the specified patched file, which can later be
 * replayed by the apply_patch procedure.
 *
 * The filename of the original file is stored in the patch and when the patch
 * is later applied, it will be applied to a file with the same name. Therefore,
 * names should be specified as relative path names.
 *
 * @param original A string containing a file name of the unpatched file.
 * @param patched A string containing a file name of the patched file.
 * @param patch A string containing a file name for the patch to be written to.
 * @param message A string containing the name of a text file which will be included as a text description (string may be empty).
 */
void create_patch (const std::string &original, const std::string &patched, const std::string &patch, const std::string &message)
{
	std::cout
		<< "Original file: " << original << std::endl
		<< "Patched file: " << patched << std::endl;

	//Attempt to open the files to compare and immediately seek to the end
	std::ifstream a (original.c_str (), std::ios::binary | std::ios::ate);
	if (!a) die (ERR_CANT_READ, "Could not open original file");
	std::ifstream b (patched.c_str (), std::ios::binary | std::ios::ate);
	if (!b) die (ERR_CANT_READ, "Could not open patched file");

	//Must be the same size
	size_t size_a = (size_t)a.tellg ();
	size_t size_b = (size_t)b.tellg ();
	if (size_a != size_b) die (ERR_SIZE_MISMATCH, "Files have different sizes");

	//Seek to the beginning
	a.seekg (0);
	b.seekg (0);

	//Don't skip whitespace
	a >> std::noskipws;
	b >> std::noskipws;

	//No differences yet
	DIFFS diffs;
	SDifference diff;

	std::cout << "Comparing files...";

	//Keep comparing bytes
	int blocks = 0;
	int bytes = 0;
	bool found = false;
	int same = 0;
	unsigned char byte_a;
	unsigned char byte_b;
	while (!a.eof ())
	{
		//Read a byte from both streams
		a >> byte_a;
		b >> byte_b;

		//Check them for being identical
		if (byte_a != byte_b)
		{
			//Different -- The first of a set?
			if (!found)
			{
				//Store offset
				diff.offset = a.tellg ();
				--diff.offset;
				found = true;
			}

			//A new byte has been found
			diff.add (byte_a, byte_b);
			same = 0;
		} else {
			//Identical -- In a set of different bytes?
			if (found)
			{
				//Add the byte
				diff.add (byte_a, byte_b);

				//If there's too many, mark the set closed
				if (++same == 8)
				{
					diffs.push_back (diff);
					found = false;

					//Update statistics
					bytes += (*diffs.rbegin ()).src.size ();
					++blocks;
				}
			}
		}
	}

	//Something left? Add it
	if (found)
	{
		diffs.push_back (diff);

		//Update statistics
		bytes += (*diffs.rbegin ()).src.size ();
		++blocks;
	}

	std::cout << " DONE" << std::endl;

	//Nothing to do?
	if (diffs.empty ()) die (ERR_IDENTICAL, "Nothing to patch, both files are identical");

	std::cout
		<< "Found "
		<< bytes
		<< " "
		<< ((bytes > 1) ? "bytes" : "byte")
		<< " in "
		<< blocks
		<< " "
		<< ((blocks > 1) ? "blocks" : "block")
		<< "."
		<< std::endl;

	//Prepare message
	std::string msg;
	if (!message.empty ())
	{
		std::ifstream m (message.c_str (), std::ios::binary | std::ios::ate);
		if (!m) die (ERR_CANT_READ, "Could not open message file");
		msg.resize (m.tellg ());
		m.seekg (0);
		m.read (&msg [0], msg.size ());
	}

	std::cout << "Writing patch file: " << patch << std::endl;

	//Attempt to create the patch file
	std::ofstream p (patch.c_str (), std::ios::binary | std::ios::trunc);
	if (!p) die (ERR_CANT_WRITE, "Can't write to patch file");

	//Simple header
	p.write ((const char *)SIGNATURE, strlen (SIGNATURE));
	//Original filename
	p << original;
	//Size of original filename
	p.write ((const char *)&size_a, sizeof (size_a));
	//User-specified message, even if empty
	p << msg;

	//Write all the differences
	DIFFS_ITOR itor = diffs.begin ();
	DIFFS_ITOR end = diffs.end ();
	while (itor != end) p << *itor++;

	//All done!
	std::cout << "Completed!" << std::endl;
}

void help ()
{
	std::cout
		<< "Patcher v1.1 (c) 2011 by Steven Don" << std::endl
		<< "===================================" << std::endl
		<< "This program creates or applies a patch to a binary file." << std::endl << std::endl
		<< "Usage:" << std::endl
		<< "\tpatcher [--original <fn> --patched <fn>] --patch <fn> [--message <fn>] [--executable] [--reverse]" << std::endl << std::endl
		<< "Parameters:" << std::endl
		<< "\t--original <fn>\tName of unpatched file to create patch from." << std::endl
		<< "\t--patched <fn>\tName of patched file to create patch from." << std::endl
		<< "\t--patch <fn>\tName of patch file to apply or create." << std::endl
		<< "\t--message <fn>\tName of file to include as message in patch." << std::endl
		<< "\t--executable\tPackage patch with executable." << std::endl
		<< "\t--reverse\tReverses (undoes) a patch during application." << std::endl
		<< "\t--help\tDisplay this message." << std::endl << std::endl
		<< "If only a patch file is specified, the patch will be applied. Otherwise, the " << std::endl
		<< "specified unpatched and patched files will be compared and a patch file will be generated from that." << std::endl << std::endl
		<< "When creating a patch, use relative (not absolute) paths for the original file " << std::endl
		<< "name." << std::endl << std::endl;

	exit (0);
}

// Entry point
////////////////////////////////////////////////////////////////////////////////

int main (int argc, char *argv[])
{
	std::string original;
	std::string patched;
	std::string patch;
	std::string message;
	std::string self;
	bool reverse = false;
	bool executable = false;

	//This can't be
	if (argc < 1) return -1;

	//Check for embedded patch
	self = get_running (argv [0]);
	patch = get_embedded (self);
	bool is_embedded = !patch.empty ();

	//No parameters
	if (argc < 2 && !is_embedded) help ();

	//Read parameters
	for (int i = 1; i < argc; ++i)
	{
		if (!is_embedded)
		{
			//Help?
			if (!strcmp (argv [i], "--help"))
			{
				help ();
			}

			//Check for original file
			if (!strcmp (argv [i], "--original"))
			{
				if (i + 1 == argc)
					die (ERR_INVALID_PARAM, "No name specified for original file");
				original = argv [++i];
				continue;
			}

			//Check for patched file
			if (!strcmp (argv [i], "--patched"))
			{
				if (i + 1 == argc)
					die (ERR_INVALID_PARAM, "No name specified for patched file");
				patched = argv [++i];
				continue;
			}

			//Check for patch file
			if (!strcmp (argv [i], "--patch"))
			{
				if (i + 1 == argc)
					die (ERR_INVALID_PARAM, "No name specified for patchfile");
				patch = argv [++i];
				continue;
			}

			//Check for message file
			if (!strcmp (argv [i], "--message"))
			{
				if (i + 1 == argc)
					die (ERR_INVALID_PARAM, "No name specified for message file");
				message = argv [++i];
				continue;
			}

			//Check to embed
			if (!strcmp (argv [i], "--executable"))
			{
				executable = true;
				continue;
			}
		}

		//Check for reversal
		if (!strcmp (argv [i], "--reverse"))
		{
			reverse = true;
			continue;
		}
	}

	//Must be a patch file
	if (patch.empty ())
		die (ERR_NO_PATCH, "No patchfile specified");

	//No original and no patched file specified? Apply patch.
	if (original.empty () && patched.empty ())
	{
		apply_patch (patch, reverse);
		if (is_embedded) remove (patch.c_str ());
		return 0;
	}

	//Either original or patched file specified? Failure.
	if (original.empty () || patched.empty ())
		die (ERR_MISSING_SOURCE, "Must specify by original and patched file");

	//Must it be embedded? Prepare a temporary filename for the patch.
	std::string output = patch;
	if (executable) patch = tmpnam (NULL);

	//Both are specified, create patch
	create_patch (original, patched, patch, message);

	//Must it be embedded? Then do so.
	if (executable)
	{
		embed_patch (self, patch, output);
		remove (patch.c_str ());
	}

	return 0;
}
